/*
Copyright (C) 2002-2003 Victor Luchits

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// snd_public.h -- sound dll information visible to engine

#define	SOUND_API_VERSION	19

//===============================================================

struct sfx_s;

//
// functions provided by the main engine
//
typedef struct
{
	// drops to console a client game error
	void			(*Error)( char *msg );

	// console messages
	void			(*Print)( char *msg );

	// dynvars
	dynvar_t	*(*Dynvar_Create)( const char *name, qboolean console, dynvar_getter_f getter, dynvar_setter_f setter );
	void		(*Dynvar_Destroy)( dynvar_t *dynvar );
	dynvar_t	*(*Dynvar_Lookup)( const char *name );
	const char	*(*Dynvar_GetName)( dynvar_t *dynvar );
	dynvar_get_status_t (*Dynvar_GetValue)( dynvar_t *dynvar,void **value );
	dynvar_set_status_t (*Dynvar_SetValue)( dynvar_t *dynvar,void *value );
	void		(*Dynvar_AddListener)( dynvar_t *dynvar, dynvar_listener_f listener );
	void		(*Dynvar_RemoveListener)( dynvar_t *dynvar, dynvar_listener_f listener );

	// console variable interaction
	cvar_t			*(*Cvar_Get)( const char *name, const char *value, int flags );
	cvar_t			*(*Cvar_Set)( const char *name, const char *value );
	void			(*Cvar_SetValue)( const char *name, float value );
	cvar_t			*(*Cvar_ForceSet)( const char *name, const char *value );	// will return 0 0 if not found
	float			(*Cvar_VariableValue)( const char *name );
	char			*(*Cvar_VariableString)( const char *name );

	int				(*Cmd_Argc)( void );
	char			*(*Cmd_Argv)( int arg );
	char			*(*Cmd_Args)( void );	// concatenation of all argv >= 1

	void			(*Cmd_AddCommand)( char *name, void(*cmd)( void ) );
	void			(*Cmd_RemoveCommand)( char *cmd_name );
	void			(*Cmd_ExecuteText)( int exec_when, const char *text );
	void			(*Cmd_Execute)( void );

	// files will be memory mapped read only
	// the returned buffer may be part of a larger pak file,
	// or a discrete file from anywhere in the quake search path
	// a -1 return means the file does not exist
	// NULL can be passed for buf to just determine existance
	int				(*FS_FOpenFile)( const char *filename, int *filenum, int mode );
	int				(*FS_Read)( void *buffer, size_t len, int file );
	int				(*FS_Write)( const void *buffer, size_t len, int file );
	int				(*FS_vPrintf)( int file, const char *format, va_list argptr ); // wsw -- PLX
	int				(*FS_Tell)( int file );
	int				(*FS_Seek)( int file, int offset, int whence );
	int				(*FS_Eof)( int file );
	int				(*FS_Flush)( int file );
	void			(*FS_FCloseFile)( int file );
	int				(*FS_GetFileList)( const char *dir, const char *extension, char *buf, size_t bufsize, int start, int end );
	int				(*FS_LoadFile)( const char *path, void **buffer, void *stack, size_t stackSize );
	void			(*FS_FreeFile)( void *buffer );

	unsigned int	(*Milliseconds)( void );
	void			(*PageInMemory)( qbyte *buffer, int size );

	// managed memory allocation
	struct mempool_s *(*Mem_AllocPool)( const char *name, const char *filename, int fileline );
	void			*(*Mem_Alloc)( struct mempool_s *pool, int size, const char *filename, int fileline );
	void			(*Mem_Free)( void *data, const char *filename, int fileline );
	void			(*Mem_FreePool)( struct mempool_s **pool, const char *filename, int fileline );
	void			(*Mem_EmptyPool)( struct mempool_s *pool, const char *filename, int fileline );

	void			(*GetEntitySpatilization)( int entnum, vec3_t origin, vec3_t velocity );

	void			*(*LoadLibrary)( const char *name, dllfunc_t *funcs );
	void			(*UnloadLibrary)( void **lib );
} sound_import_t;

//
// functions exported by the client game subsystem
//
typedef struct
{
	// if API is different, the dll cannot be used
	int				(*API)( void );

	// the init function will be called at each restart
	qboolean		(*Init)( void );

	void			(*Shutdown)( void );

	void			(*SoundsInMemory)( void );
	void			(*FreeSounds)( void );
	void			(*StopAllSounds)( void );

	void			(*Clear)( void );
	void			(*Update)( const vec3_t origin, const vec3_t velocity, const vec3_t v_forward, const vec3_t v_right, const vec3_t v_up );
	void			(*Activate)( qboolean active );

	// playing
	struct sfx_s	*(*RegisterSound)( const char *sample );
	void			(*FreeSound)( struct sfx_s *sfx );

	void			(*StartFixedSound)( struct sfx_s *sfx, const vec3_t origin, int channel, float fvol, float attenuation );
	void			(*StartRelativeSound)( struct sfx_s *sfx, int entnum, int channel, float fvol, float attenuation );
	void			(*StartGlobalSound)( struct sfx_s *sfx, int channel, float fvol );

	void			(*StartLocalSound)( const char *s );

	void			(*AddLoopSound)( struct sfx_s *sfx, int entnum, float fvol, float attenuation );

	// cinema
	void			(*RawSamples)( int samples, int rate, int width, int channels, const qbyte *data, qboolean music );

	// music
	void			(*StartBackgroundTrack)( const char *intro, const char *loop );
	void			(*StopBackgroundTrack)( void );
} sound_export_t;
